﻿using System;
using System.Diagnostics.Tracing;
using System.Linq;
using System.Reflection;

class Program
{
    static MyEventListener _listener;

    static Program()
    {
        _listener = new MyEventListener(DotnetRuntimeKeyword.EXCEPTIONKEYWORD, new PrimitiveEnum[] { ExceptionEvent.ExceptionCatchStart, ExceptionEvent.ExceptionCatchStop });

        // _listener = new MyEventListener(DotnetRuntimeKeyword.EXCEPTIONKEYWORD);
    }

    static void Main(string[] args)
    {
        Console.WriteLine(Environment.Version);
        Console.WriteLine("In-proc ETW Enabled on .NET Core 2.2 or later");

        while (true)
        {
            string result = Console.ReadLine();

            /*
            if (result == "s")
            {
                if (_listener != null)
                {
                    _listener.Dispose();
                }

                _listener = new MyEventListener(DotnetRuntimeKeyword.EXCEPTIONKEYWORD, new PrimitiveEnum[] { ExceptionEvent.ExceptionCatchStart, ExceptionEvent.ExceptionCatchStop });
            }
            */

            try
            {
                throw new ApplicationException("Exception-thrown");
            }
            catch (Exception)
            {
                Console.WriteLine($"{DateTime.Now}: User-exception");
            }

            GC.Collect(GC.MaxGeneration, GCCollectionMode.Forced);
        }
    }
}

public enum DotnetRuntimeKeyword : long
{
    GCKEYWORD = 0x01,
    EXCEPTIONKEYWORD = 0x00008000
}

public enum ExceptionEvent
{
    ExceptionThrown_V1 = 80,
    ExceptionCatchStart = 250,
    ExceptionCatchStop = 251,
    ExceptionFinallyStart = 252,
    ExceptionFinallyStop = 253,
    ExceptionFilterStart = 254,
    ExceptionFilterStop = 255,
    ExceptionThrownStop = 256,
}

internal class MyEventListener : EventListener
{
    // https://docs.microsoft.com/en-us/dotnet/framework/performance/clr-etw-keywords-and-levels

    EventSource _eventSource;
    long _filterKeyword;
    int[] _filterEventId;

    public MyEventListener(DotnetRuntimeKeyword filterKeyword, PrimitiveEnum[] filterEventId = null)
    {
        this._filterKeyword = (long)filterKeyword;
        this._filterEventId = filterEventId?.Select((elem) => (int)elem).ToArray();
        this.EventSourceCreated += MyEventListener_EventSourceCreated;
    }

    private void MyEventListener_EventSourceCreated(object sender, EventSourceCreatedEventArgs e)
    {
        if (e.EventSource.Name == "Microsoft-Windows-DotNETRuntime")
        {
            _eventSource = e.EventSource;
            e.EventSource.EventCommandExecuted += EventSource_EventCommandExecuted;
            EnableEvents(e.EventSource, EventLevel.Informational, (EventKeywords)this._filterKeyword);
        }
    }

    private void EventSource_EventCommandExecuted(object sender, EventCommandEventArgs e)
    {
        if (e.Command == EventCommand.Enable)
        {
            // e.DisableEvent((int)ExceptionEvent.ExceptionThrown_V1);
            // e.DisableEvent((int)ExceptionEvent.ExceptionCatchStart);
            // e.DisableEvent((int)ExceptionEvent.ExceptionCatchStop);
            // e.DisableEvent((int)ExceptionEvent.ExceptionFinallyStart);
            // e.DisableEvent((int)ExceptionEvent.ExceptionFinallyStop);
            // e.DisableEvent((int)ExceptionEvent.ExceptionFilterStart);
            // e.DisableEvent((int)ExceptionEvent.ExceptionFilterStop);
            // e.DisableEvent((int)ExceptionEvent.ExceptionThrownStop);

            FilterEventId(_eventSource, e);
        }
    }

    private void FilterEventId(EventSource eventSource, EventCommandEventArgs e)
    {
        Type type = eventSource.GetType();
        FieldInfo fi = type.GetField("m_eventData", System.Reflection.BindingFlags.NonPublic | System.Reflection.BindingFlags.Instance);
        if (fi != null)
        {
            object objMetadataArray = fi.GetValue(eventSource); // System.Diagnostics.Tracing.EventSource.EventMetadata []
            System.Collections.IEnumerable enumerator = objMetadataArray as System.Collections.IEnumerable;

            foreach (object objItem in enumerator)
            {
                object descriptor = GetDescriptor(objItem);
                if (descriptor == null)
                {
                    continue;
                }

                (long keyword, int eventId) = GetKeywordAndEventId(descriptor);

                if (((long)keyword & this._filterKeyword) == this._filterKeyword)
                {
                    if (_filterEventId?.Contains(eventId) == false)
                    {
                        e.DisableEvent(eventId);
                    }
                }
            }
        }
    }

    private (long, int) GetKeywordAndEventId(object descriptor)
    {
        object keyword = GetPropertyReflect(descriptor, "Keywords");
        object eventValue = GetPropertyReflect(descriptor, "EventId");

        if (keyword == null || eventValue == null)
        {
            return (0, 0);
        }

        return ((long)keyword, (int)eventValue);
    }

    private object GetDescriptor(object objItem)
    {
        return GetFieldReflect(objItem, "Descriptor");
    }

    object GetPropertyReflect(object obj, string propertyName)
    {
        Type type = obj.GetType();
        PropertyInfo pi = type.GetProperty(propertyName, BindingFlags.Public | BindingFlags.Instance);
        if (pi != null)
        {
            return pi.GetValue(obj);
        }

        return null;
    }

    object GetFieldReflect(object obj, string fieldName)
    {
        Type type = obj.GetType();
        FieldInfo fi = type.GetField(fieldName, BindingFlags.Public | BindingFlags.Instance);
        if (fi != null)
        {
            return fi.GetValue(obj);
        }

        return null;
    }

    protected override void OnEventWritten(EventWrittenEventArgs eventData)
    {
        if (((long)eventData.Keywords & _filterKeyword) == _filterKeyword)
        {
            Console.WriteLine($"{DateTime.Now}: {eventData.EventName}");
        }
    }
}